#include <stdio.h>
#include <FSextend.h>
#include <FSopenc.h>
#include <mysql.h>
#include <stdlib.h>
#include "sqlmy.h"
#include "dbf.h"

char errores[100];

/*************************************************************/
/* OPEN a MySQL database
 * A UDF to connect to a MySQL database
 *
 * PARAMETERS : One is required - the name of the database
 *
 * open_my(db, host, user, pass, port, unixsock, flags)       */

FSudfname( open_my)
{
  char *db_name, *my_host, *my_user, *my_pass, *my_unixsock;
  int my_port, my_flags;

  int i;

  FSinit();
  err_log = getenv("MY_ERROR_LOG");
  verbosity = atoi(getenv("MY_LEVEL_VERBOSITY"));
  /*if (PCOUNT != 1 || _parinfo (1) != CHARACTER)*/
  if (PCOUNT < 1 || _parinfo (1) != CHARACTER)
    {
      pr_error("open_my() requires at least db name");
      _retni (ERR_WRONG_NUM_ARGUMENTS);
      FSreturn;
    }

/* Verify the number of arguments */

  if (PCOUNT > 7) {
    pr_error("The number of argument is exceded");
    _retni(ERR_WRONG_NUM_ARGUMENTS);
    FSreturn;
  }

/* initialize the variable to standard values  */

  my_host = "localhost";
  my_user = "";
  my_pass = "";
  my_port = 3306;
  my_unixsock = NULL;
  my_flags = 0;

/* set the variable depending on the arguments pass it */

  for(i=1;i<=PCOUNT;i++){
    switch (i) {
      case 1: /* database name */
        db_name= _parc(i);
        break;
      case 2:  /* user name */
        my_host= _parc(i);
        break;
      case 3:  /* user name */
        my_user= _parc(i);
        break;
      case 4:  /* password if set, if not null */
        my_pass= _parc(i);
        break;
      case 5:  /* port number */
        my_port= _parni(i);
        break;
      case 6:  /* unix socket file name */
        my_unixsock= _parc(i);
        break;
      case 7:  /* flags */
        my_flags= _parni(7);
        break;
    }
  }

  /* only one open connection is allowed per session */
  if (IsOpen) {
    pr_error("Attempt to open a database twice.");
    _retni(ERR_OPEN_TWICE);
    FSreturn;
  }

  if(!mysql_init(&conn)) {
    pr_error("Cannot create the conection structure.");
    _retni(ERR_CANNOT_CREATE_STRUCTURE);
    FSreturn;
  }

  if(!mysql_real_connect(&conn,my_host,my_user,my_pass,db_name,my_port,my_unixsock,my_flags)){
    verifica(&conn);
    sprintf(errores, "User %s cannot connect to %s host, database %s" , my_user,my_host,db_name);
    pr_error(errores);
    _retni(ERR_CANNOT_CONNECT);
    FSreturn;
  }
  if (verbosity > 2){
    sprintf(errores, "User %s opened : %s db", getenv("USER"),conn.db);
    pr_error(errores);
  }
  IsOpen=1;
  _retni (0);
  FSreturn;
}

/*************************************************************/
/* CLOSE a My-SQL database */
FSudfname( close_my)
{
  FSinit();
  if(IsOpen==0){
    sprintf(errores, "There isn't actually any database open");
    pr_error(errores);
    _retni(ERR_NO_DB_OPEN);
    FSreturn;
  }
  err_log = getenv("MY_ERROR_LOG");
  verbosity = atoi(getenv("MY_LEVEL_VERBOSITY"));
  if (verbosity > 2){
    sprintf(errores, "User %s closed : %s db\n", getenv("USER"),conn.db);
    pr_error(errores);
  }
  mysql_close(&conn);
  IsOpen = 0;
  FSreturn;
}


/*************************************************************/
/* Ask if still is alive a connection with MYSQL database */
FSudfname( alive_my )
{
  FSinit();
  err_log = getenv("MY_ERROR_LOG");
  verbosity = atoi(getenv("MY_LEVEL_VERBOSITY"));
  if (verbosity > 2){
    sprintf(errores, "User %s is asking if %s db is alive", getenv("USER"),conn.db);
    pr_error(errores);
  }
  if(!mysql_ping(&conn)){
    sprintf(errores,"Error: %s",mysql_error(&conn));
    pr_error(errores);
    _retni(ERR_CONNECTION_LOST);
    FSreturn;
  }
  else {
    _retni(1);
    FSreturn;
  }
}


/*************************************************************/
/* RETRIEVE tuples from a My-SQL database 
 *
 *   Four parameters are required.
 *
 *      type_of_retrieve              input parameter
 *                       0: Retrieve the result row-by-row
 *                       1: Retrieve the entire result
 *
 *      select_statement              input parameter
 *
 *      pointer to temp_file_name     output parameter
 *
 *      pointer to number_of_fields   output parameter
 *
 *   Return value : number_of_tuples
 *
 */
FSudfname( ret_my )
{
  char *sel_statement, *retrieve_dir, *err_log;
  char cmd[3000];
  unsigned char garbage[5];
  char tfilename[80];
  int nFields, nRows, i, j, pid, type_of_retrieve;
  static int RelCount = 0;
  FILE *tfp;
  char *ctime();
  long time(), now;
  MYSQL_RES *result; 
  MYSQL_ROW tuple;
  MYSQL_FIELD *fields;

  FSinit();

  err_log = getenv("MY_ERROR_LOG");
  verbosity = atoi(getenv("MY_LEVEL_VERBOSITY"));
  if (PCOUNT != 4 || _parinfo(1) != CHARACTER)
    {
      pr_error("ret_my() requires three parameters");
      _retni(ERR_WRONG_NUM_ARGUMENTS);
      FSreturn;
    }

  sel_statement = _parc(1);
  type_of_retrieve = _parni(2);
  if(verbosity>2){
    sprintf(errores, "User %s executed : %s db\n", getenv("USER"),sel_statement);
    pr_error(errores);
  }

  mysql_query(&conn,sel_statement);
  if (verifica(&conn)){
    _retni(ERR_CANNOT_EXEC_SELECT);
    FSreturn;
  }

  /* verify the type of retrieve*/

  if (type_of_retrieve){
	 result=mysql_store_result(&conn);
  }
  else {
	result=mysql_use_result(&conn);
  }

  if (verifica(&conn)){
    _retni(ERR_CANNOT_RECOVERY_RESULT);

    FSreturn;
  }
  nFields=mysql_num_fields(result);


  /* store number of fields in third parameter */
  _storni( (long)nFields, 4 );

  /* make temp retrieved-records file */
  retrieve_dir = getenv("MY_RETRIEVE_DIR");
  if (retrieve_dir == NULL) retrieve_dir="";
  pid = getpid();
#ifdef TMP_FILE_TXT
  sprintf( tfilename, "%s%s%d%c%d", retrieve_dir, "myselect",
  ++RelCount, '.', pid); 
#else
  sprintf( tfilename, "%s%s%d%d%c%s", retrieve_dir, "myselect",
  ++RelCount,pid, '.', "dbf"); 
#endif
  _storc(tfilename,3);

  tfp = fopen( tfilename, "w" );
  if (tfp == NULL){
    pr_error("unable to open tmp file in ret_my()");
    _retni(ERR_CANNOT_OPEN_TMP_FILE);
    FSreturn;
  }

  fields=mysql_fetch_fields(result);
#ifdef TMP_FILE_TXT
  nRows=write_in_txt(tfp,result);
#else
  nRows=write_in_dbf(tfp,result);
#endif
  fclose( tfp );  /* close the temporary text file */
  mysql_free_result(result); /* frees the memory asociated with result */
  if(nRows<0) { 
    _retni(nRows); /* Return an error code */
    FSreturn;
  }

 
  /*We put in the structure the number of Rows */
  tfp = fopen( tfilename, "r+" );
  if (tfp == NULL){
    pr_error("unable to open tmp file in ret_my()");
    _retni(ERR_CANNOT_OPEN_TMP_FILE);
    FSreturn;
  }
  _retni( nRows );   /* return number of records retrieved */
  FSreturn;
}

/*Function: verifica(), verify if the excecuted SQL instruction
  it's OK. If true return 0, else return -1 recording the problem
  in its error log, if any. It's up to the programmer abort or not
  when status is -1 is received */

int verifica(MYSQL *mysql){
  if(mysql_errno(mysql)){
    sprintf(errores,"Error: %s",mysql_error(mysql));
    pr_error(errores);
  }
  return;
}

/*************************************************************/
/* RETRIEVE tuples from a My-SQL database 
 *
 *      select_statement              input parameter
 *
 *   Return value : pointer to a aRecord data structure
 *
	Information structure (aRecord)
	=====================

     --------------------------------
     | 1 | 2 | 3 | 4 | ....|nFields |
     --------------------------------
       |
       v
       ---------------------------------------
       |name_Field_1|type_Field_1|           |
       ---------------------------------------
                                   |
                                   v
                      --------------------------------------------------
                      | value_1_of_field_1 |..|  value_nRows_of_field_1|
                      --------------------------------------------------


  aRecord is an array in FlagShip land, but come from C land. I mean
   a C function build this array.

  If we crete aRecord from FlagShip land directly. We should do:

  aRecord := {
              {name_1,type_1,{value_1, value_2,..,value_nRows}},
              {name_2,type_2,{value_1,value_2,..,value_nRows}},..,
              {name_nFields,type_nFields,{value_1, value_2,..,
              value_nRows}}
             }

   The name of the first field is:          aRecord[1,1]
   The type of the first field is:          aRecord[1,2]
   The first value of the first field is :  aRecord[1,3,1]
   The third value of the second field is :  aRecord[2,3,3]

 As you can see, the second index of the array remain the same (3),
 when you recover a value from the array aRecord from the FlagShip
 land. The first index refer to field, and the third is a value itself.

*/

   
UDF_DECL( ret_mem_my )   /* max 10 chars in lowercase */
{

  FSvar *varNested, *dummy, *retVal=VAR_NEW; 

  char *sel_statement, *retrieve_dir, *err_log;
  int nFields, nRows, iFields, jRows;
  char *fieldType;
  char *ctime();
  long time(), now;
  MYSQL_RES *result; 
  MYSQL_ROW tuple;
  MYSQL_FIELD *fields;
  char width[15];
  char formato[25];
  char value[75];


  err_log = getenv("MY_ERROR_LOG");
  verbosity = atoi(getenv("MY_LEVEL_VERBOSITY"));
   
  sel_statement = _parc(1);
  if(verbosity>2){
    sprintf(errores, "User %s executed : %s", getenv("USER"),sel_statement);
    pr_error(errores);
  }

  if(mysql_query(&conn,sel_statement)){
    sprintf(errores,"Error: %s",mysql_error(&conn));
    pr_error(errores);
    SET_VAR_INT(retVal,ERR_CANNOT_EXEC_SELECT);
    return retVal;
  }

  if(!(result=mysql_store_result(&conn))){
    sprintf(errores,"Error: %s",mysql_error(&conn));
    pr_error(errores);
    SET_VAR_INT(retVal,ERR_CANNOT_RECOVERY_RESULT);
    return retVal;
  }
  nFields=mysql_num_fields(result);
 
  if(( fieldType = (char *) malloc( nFields )) == NULL ) {
      pr_error(DB_MEMORY_ERROR);
      SET_VAR_INT(retVal,ERR_CANNOT_ALLOCATE_MEM);
      return retVal;
  }

  nRows=mysql_num_rows(result);

  /* Verify if we have exceeded the number of array elements (65535) */
  if (nFields + nFields * 3 + nFields * nRows > 65535) {
    sprintf(errores,"The number of elements exceed 65535");
    pr_error(errores);
    SET_VAR_INT(retVal,ERR_EXCEED_NUM_ARRY_ELEM);
    free(fieldType);
    return retVal;
  }

  /* aRecord data structure creation  */

  varNested = VAR_NEW_ARRAY(dummy,nFields);
  for(iFields=1;iFields<=nFields;iFields++){
    SET_VAR_COPY(VAR_ARRELEM(varNested,iFields),VAR_NEW_ARRAY(dummy,3));
    SET_VAR_COPY(VAR_ARRELEM(VAR_ARRELEM(varNested,iFields),3),
      VAR_NEW_ARRAY(dummy,nRows));
  }

  /* End data structure creation */

  /* Filling the correct field name and type to its place */
 
  for(iFields=0;iFields<nFields;iFields++) 
  {
    SET_VAR_CHR(VAR_ARRELEM(VAR_ARRELEM(varNested,iFields+1),1),
                result->fields[iFields].name);

   /* Now filling the type of field */
    switch(result->fields[iFields].type)
    {
      case FIELD_TYPE_TINY:
      case FIELD_TYPE_SHORT:
      case FIELD_TYPE_LONG:
      case FIELD_TYPE_LONGLONG:
      case FIELD_TYPE_INT24:
      case FIELD_TYPE_ENUM:
      case FIELD_TYPE_SET:
        SET_VAR_CHR(VAR_ARRELEM(VAR_ARRELEM(varNested,iFields+1),2),"I");
        fieldType[iFields]='I';
	break;
      case FIELD_TYPE_TIMESTAMP:
      case FIELD_TYPE_TIME:
      case FIELD_TYPE_DATETIME:
      case FIELD_TYPE_TINY_BLOB:
      case FIELD_TYPE_STRING:
      case FIELD_TYPE_YEAR:
      case FIELD_TYPE_VAR_STRING:
        SET_VAR_CHR(VAR_ARRELEM(VAR_ARRELEM(varNested,iFields+1),2),"C");
        fieldType[iFields]='C';
	break;
      case FIELD_TYPE_DATE:
        SET_VAR_CHR(VAR_ARRELEM(VAR_ARRELEM(varNested,iFields+1),2),"D");
        fieldType[iFields]='D';
	break;
      case FIELD_TYPE_FLOAT:
      case FIELD_TYPE_DECIMAL:
      case FIELD_TYPE_DOUBLE:
        SET_VAR_CHR(VAR_ARRELEM(VAR_ARRELEM(varNested,iFields+1),2),"N");
        fieldType[iFields]='N';
	break;

  /* The next type of field are not supported yet, if you really
     want to use it, please let me know, because I am thinking in
     supporting it */
      case FIELD_TYPE_NULL:
      case FIELD_TYPE_NEWDATE:
      case FIELD_TYPE_MEDIUM_BLOB:
      case FIELD_TYPE_LONG_BLOB:
      case FIELD_TYPE_BLOB:
		pr_error("This type of data is not supported yet");
                SET_VAR_INT(retVal,ERR_DATA_NOT_SUPPORTED);
                free(fieldType);
                return retVal;
    } /* End filling the type of field */
  }   /* End of filling the field name and type to its place */

  /* Filling the structure with its data  */

  nRows=0;
  while ((tuple=mysql_fetch_row(result)))
  {
    nRows++;
    for(iFields=0;iFields<nFields;iFields++) {
      switch(fieldType[iFields]){
        case 'I':
          formato[0]='\0';
          sprintf(width,"%d",result->fields[iFields].length);
          strcat(formato,"%"); 
          strcat(formato,width); strcat(formato,"s");
          sprintf(value, formato, tuple[iFields]);
          SET_VAR_INT(VAR_ARRELEM(VAR_ARRELEM(
            VAR_ARRELEM(varNested,iFields+1),3),nRows), atol(value));
          break;
        case 'C':
          SET_VAR_CHR(VAR_ARRELEM(VAR_ARRELEM(
            VAR_ARRELEM(varNested,iFields+1),3),nRows),tuple[iFields]);
          break;
        case 'D':
          SET_VAR_CHR(VAR_ARRELEM(VAR_ARRELEM(
            VAR_ARRELEM(varNested,iFields+1),3),nRows),tuple[iFields]);
          /*SET_VAR_DATE(VAR_ARRELEM(VAR_ARRELEM(
            VAR_ARRELEM(varNested,iFields+1),3),nRows),tuple[iFields]);*/
          break;
        case 'N':
          formato[0]='\0';
          sprintf(width,"%d",result->fields[iFields].length);
          strcat(formato,"%"); 
          strcat(formato,width); strcat(formato,"s");
          sprintf(value, formato, tuple[iFields]);
          SET_VAR_NUM(VAR_ARRELEM(VAR_ARRELEM(
            VAR_ARRELEM(varNested,iFields+1),3),nRows),atof(value));
          break;
      }
    }
  }
  free(fieldType);
  return varNested;
}



/*************************************************************/
/* EXEC a My-SQL SQL command
 *
 * may be used for all SQL commands execept SELECT
 *
 * PARAMETERS : one required : the SQL string
 *
 */
FSudfname( exec_my )
{
  char *cmd;
  char *ctime(), *err_log;
  long time(), now, affected_rows;

  FSinit();
  err_log = getenv("MY_ERROR_LOG");
  verbosity = atoi(getenv("MY_LEVEL_VERBOSITY"));

  if (PCOUNT != 1 || _parinfo (1) != CHARACTER)
    {
      pr_error("exec_my() requires a parameter");
      _retni (ERR_WRONG_NUM_ARGUMENTS);
      return;
    }

  cmd = _parc(1);  

  mysql_query(&conn,cmd);
  if (verifica(&conn)){
    _retni(ERR_EXECUTING_NO_SELECT);
    FSreturn;
  }
  affected_rows=mysql_affected_rows(&conn);

  if (verbosity > 2){
    sprintf(errores, "User %s executed : %s db", getenv("USER"),cmd);
    pr_error(errores);
  }

  _retni (affected_rows);
  FSreturn;
}
int write_in_txt(FILE *tfp, MYSQL_RES *result)
{
  int nFields, nRows,j;
  MYSQL_ROW  tuple;
  unsigned char i=0;

  nRows=0;
  nFields=mysql_num_fields(result);
  for(j=0;j<nFields;j++) fprintf(tfp, "%d |",
           result->fields[j].type);
  while ((tuple=mysql_fetch_row(result))){
      
    for(j=0;j<nFields;j++) fprintf(tfp, "%s|", tuple[j]);
    fprintf(tfp, "\n");
    nRows++;
  }
  if(!mysql_eof(result)) {  // mysql_fetch_row() failed due to an error
    sprintf(errores, "Error: %s", mysql_error(&conn));
    pr_error(errores);
    nRows=ERR_CANNOT_FETCH_ROWS;
  }
  return nRows;
}
